! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

#ifdef VTRACE
#include "vt_user.inc"
#else
#define VT_USER_START(n)
#define VT_USER_END(n)
#endif

!--------------------------------------------------------------!
!                                                              !
!            Coupling of FD4 to other parallel model           !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Coupling of FD4 to other parallel model.
!!
!! *Limitations*:
!!  - sequential coupling only (i.e. FD4 operates on same processes
!!    as the coupled model)
!!  - user must ensure that couple arrays do not overlap, otherwise same data
!!    will be transferred multiple times. Problematic when using $fd4_couple_put!
!!
!! Preprocessor options:
!!  - $FD4_COUPLE_DEBUG: verbose debug output to fort.* files
!!  - $FD4_COUPLE_FACE_ARRAYS_OVERLAP: tell FD4 whether the couple arrays for face variables overlap
!!    at the inner boundaries (one cell in face direction)
!!  - $FD4_COUPLE_NEIGHBOR_FACE_CHECK: for $fd4_couple_get. Checks for face variables if the neighbor block exists.
!!    Prepares data types depending on block allocation status of blocks! In practice, $fd4_couple_create
!!    would be necessary after partitioning has changed. But this is currently the only way to $get face variables
!!    correctly at block - no-block boundaries (because the default is to get upper face from the next block,
!!    which does not work, if the next block is not allocated)
!!
!! Performance Notes:
!!  - The fd4_couple_desc array should be sorted firstly by data field and secondly by rank, e.g.
!!    exactly one continuous section of fields for each rank
!!
!! Some notes about sorting of the components of the mpi datatypes we are building here:
!!  - components are identified by position and variable/step
!!  - sorting is required for correct matching on sender and receiver side
!!  - sorting happens only for the position (in $fd4_mpitype_commit)
!!  - variable/step ordering is given by the order in the array description $desc
!!  - that means
!!    - sorting in $fd4_mpitype_commit has to be stable!
!!    - *all processes put the arrays in the same order into* $desc
!!
!! *TODO*:
!!  - implement a routine which checks the couple description for overlapping arrays
!!    (which are not allowed for $fd4_couple_put!)
!!     - this routine should be called in couple_create
!!     - it should disable $fd4_couple_put if necessary
!!  - solve $FD4_COUPLE_NEIGHBOR_FACE_CHECK problem: build both possible types and decide in $fd4_couple_get
!!    which type to take
!!  - provide routines optimized for 2D-partitioning (x,y)? - less intersections to calculate
!!  - hierarchical description of couple arrays: multiple variables for each couple array section (bounds)
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_couple_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use fd4_mpitype_mod
  use fd4_comm_mod
  use stack_mod
  use rbtree_int_mod
  use rbtree_int8_mod
  use section_mod

  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  
  ! public types
  public :: fd4_couple
  public :: fd4_couple_desc
  public :: fd4_domain
  public :: fd4_mpitype
  ! public subroutines
  public :: fd4_couple_create
  public :: fd4_couple_delete
  public :: fd4_couple_put
  public :: fd4_couple_get  
  public :: fd4_couple_mark_blocks

  
  integer(i_k), parameter :: CPL_NOFACE=1, CPL_FACE=2, CPL_NO=3
  integer(i_k), parameter :: CPL_GET=1, CPL_PUT=2
  

  !! Description of one coupling array provided by user.
  !!
  !!  - $rank, $idx, $st, and $abnd must be given for all blocks
  !!  - the following components must be set for local blocks:
  !!    - 3D data: $array and $offset
  !!    - 4D data: $array4, $offset, and $offset4
  !!    - $array and $array4 must not be set at the same time
  type fd4_couple_desc
    integer(i_k) :: rank                           !! owner of this block
    integer(i_k) :: idx                            !! field index in vartab
    integer(i_k) :: st                             !! time step index
    integer(i_k) :: abnd(3,2)                      !! bounds of the block in global coordinates
    real(r_k), pointer :: array(:,:,:) => NULL()   !! data array
    real(r_k), pointer :: array4(:,:,:,:) => NULL()!! 4-dimensional data array (for variables with bins)
    integer(i_k) :: offset(3) = (/1,1,1/)          !! spatial offset of the data in array, e.g. (/1,1,1/)
    integer(i_k) :: offset4 = 1                    !! offset of the data in the 4th dimension
  end type fd4_couple_desc
  
  
  !! The FD4 couple context.
  !!
  type fd4_couple
    type(fd4_domain), pointer :: domain => NULL()  !! the domain
    integer(i_k) :: loopbnd(3,2)                   !! bounds for the loop over blocks    
    integer, pointer :: arraytypes(:,:,:,:,:)      !! pre-built mpi types for couple arrays
    type(fd4_mpitype), pointer :: commtype(:)      !! couple array mpi types for each dest rank
    integer(i_k) :: toff                           !! offset of types for send/receive in commtype array
    integer(i_k) :: ncplarrays                     !! number of local couple arrays in couple description
    integer(i_k), pointer :: rankbnd(:,:,:)        !! union bounds of all couple arrays for each rank
    integer(i_k) :: faceoffset(3)                  !! offset to local block bounds union due to face variables
    integer(i_k), pointer :: mycplarrays(:)        !! indexes of my couple arrays in $fd4_cpl_desc array
    integer(i_k), pointer :: cpldescbnd(:,:)       !! first and last index of couple arrays in $fd4_cpl_desc per rank
  end type fd4_couple

!#define FD4_COUPLE_DEBUG
  
contains  


  include 'util/stack.incf'
  include 'framework/fd4_domain.incf'
  
  
  
  !! Create FD4 couple context.
  !!
  !! *TODO*:
  !!  - is it correct only to check for abnd size < 1 and not abnd size <=1 ?
  !!
  subroutine fd4_couple_create(cpl, domain, desc, err)
    type(fd4_couple),     intent(inout)        :: cpl          !! couple context
    type(fd4_domain),     intent(inout),target :: domain       !! the domain
    type(fd4_couple_desc),intent(in),target    :: desc(:)      !! description of sender arrays
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: ndesc
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: idx, idesc, dest
    integer(i_k) :: bbnd(3,2),face,intersec(3,2)
    integer(i_k) :: i,j,k
    real(r_k), pointer :: myarray(:,:,:)
    integer(i_k), pointer :: myabnd(:,:)
    integer :: blocktype
    integer(i_k) :: dim, icplarrays, facetype
    logical :: facevar_present(0:3)
#ifdef FD4_COUPLE_NEIGHBOR_FACE_CHECK
    integer(i_k) :: pos(3)
#endif
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: fext(3)
#endif

    FD4_L1_ASSERT(associated(cpl%domain),'cpl not empty')
    
    ndesc = size(desc)
    
    cpl%domain => domain
    cpl%ncplarrays = 0
    
    mpi => domain%mpi
    err = -1
    intersec(:,1) = 0
    intersec(:,2) = -1
    facevar_present = .false.
    cpl%faceoffset(1:3) = 0
    
    allocate(cpl%cpldescbnd(2,0:mpi%nprocs-1))
    cpl%cpldescbnd = -1
        

#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_create ',ndesc
#endif

    ! check params
    ! and find couple array with min address
    do idesc=1,ndesc
      idx = desc(idesc)%idx
      if(desc(idesc)%rank == mpi%rank) then
        cpl%ncplarrays = cpl%ncplarrays + 1
        if( domain%vartab(idx)%nbins == 1 .and. .not. associated(desc(idesc)%array)) then
          FD4_L1_ERROR('desc%array not associated at '//i2s(idesc))
        end if
        if( domain%vartab(idx)%nbins >  1 .and. .not. associated(desc(idesc)%array4)) then
          FD4_L1_ERROR('desc%array4 not associated at '//i2s(idesc))
        end if
        if(associated(desc(idesc)%array) .and. associated(desc(idesc)%array4)) then
          FD4_L1_ERROR('desc%array and desc%array4 associated at '//i2s(idesc))
        end if
        myarray => desc(idesc)%array
        if( domain%vartab(idx)%nbins > 1 ) myarray => desc(idesc)%array4(lbound(desc(idesc)%array4,1),:,:,:)
        ! check if abnd size < 1
        if( ANY(desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) < 0 ) ) then
        ! can't remember why I checked for <=0...
        !if( ANY(desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) <= 0 ) ) then
          FD4_L1_ERROR('desc%abnd size is <1 at '//i2s(idesc))
        end if
#if ( FD4_VERBOSE_LEVEL >= 3 )
        ! check array size for 4th dimension
        if( domain%vartab(idx)%nbins > 1 ) then
          if( lbound(desc(idesc)%array4,1) > desc(idesc)%offset4 ) then
            FD4_L1_ERROR('desc%offset4 too small at '//i2s(idesc))
          end if
          if( ubound(desc(idesc)%array4,1) < desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1 ) then
            FD4_L1_ERROR('desc%array4 too small at '//i2s(idesc))
          end if
        end if
        ! check array size for spatial dimensions
        fext = lbound(myarray)
        if( ANY( fext > desc(idesc)%offset ) ) then
          FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
        end if
        fext = ubound(myarray)
        if( ANY( fext < desc(idesc)%offset + desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) ) ) then
          FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
        end if
#endif
        ! get bounds
        if( intersec(1,1) > intersec(1,2) ) then
          ! 1st time here
          intersec = desc(idesc)%abnd
        else
          ! add this coupling array to the union of all coupling arrays
          ! (yes, variable is called intersection here...)
          intersec(1:3,1)=MIN(intersec(1:3,1),desc(idesc)%abnd(1:3,1))
          intersec(1:3,2)=MAX(intersec(1:3,2),desc(idesc)%abnd(1:3,2))
        end if
      end if
      facevar_present(domain%vartab(idx)%facevar) = .true.
      if(domain%vartab(idx)%facevar>0 .and. domain%vartab(idx)%facevar<4) then
        cpl%faceoffset(domain%vartab(idx)%facevar) = 1
      end if
      ! fill cpldescbnd
      if(cpl%cpldescbnd(1,desc(idesc)%rank)<0) then
        cpl%cpldescbnd(1,desc(idesc)%rank) = idesc
        cpl%cpldescbnd(2,desc(idesc)%rank) = idesc
      else
        cpl%cpldescbnd(2,desc(idesc)%rank) = idesc
      end if
    end do
    
    
    ! reduce the number of blocks to loop by taking the
    ! union of all coupling arrays into account
    ! 1. intersect this union with the domain size
    intersec(1:3,1)=MAX(intersec(1:3,1),domain%dbnd(1:3,1))
    intersec(1:3,2)=MIN(intersec(1:3,2),domain%dbnd(1:3,2))
    ! 2. now calculate the loop bounds
    do dim=1,3
      ! try to guess the starting block
      i = ( domain%blnum(dim) * (intersec(dim,1)-domain%dbnd(dim,1)+1) ) / (domain%dbnd(dim,2)-domain%dbnd(dim,1)+1)
      i = MIN(MAX(1,i),domain%blnum(dim))
      ! find real start
      do
        if(domain%d(dim)%bl(i)>intersec(dim,1)) then
          i=i-1
        else if(domain%d(dim)%bl(i+1)<=intersec(dim,1)) then
          i=i+1
        else
          exit
        end if
      end do
      ! shift the start forward if
      ! * facevar present 
      ! * block start and intersection start exactly at the same position 
      if(facevar_present(dim) .and. intersec(dim,1)==domain%d(dim)%bl(i) .and. i>1) i=i-1
      cpl%loopbnd(dim,1) = i 
      ! try to guess the end block
      i = ( domain%blnum(dim) * (intersec(dim,2)-domain%dbnd(dim,1)+1) ) / (domain%dbnd(dim,2)-domain%dbnd(dim,1)+1)
      i = MIN(MAX(1,i),domain%blnum(dim))
      ! find real end
      do
        if(domain%d(dim)%bl(i)>intersec(dim,2)) then
          i=i-1
        else if(domain%d(dim)%bl(i+1)<=intersec(dim,2)) then
          i=i+1
        else
          exit
        end if
      end do    
      cpl%loopbnd(dim,2) = i
    end do

    cpl%toff = 0
#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    cpl%toff = mpi%nprocs
#endif
    allocate(cpl%commtype(0:mpi%nprocs-1+cpl%toff))
    
    allocate(cpl%arraytypes(cpl%ncplarrays, cpl%loopbnd(1,1):cpl%loopbnd(1,2), &
                                            cpl%loopbnd(2,1):cpl%loopbnd(2,2), &
                                            cpl%loopbnd(3,1):cpl%loopbnd(3,2),   2  ) )

    allocate(cpl%mycplarrays(cpl%ncplarrays))
    
    cpl%arraytypes = MPI_DATATYPE_NULL

    !! pre-build types for the coupling array (subsections which correspond to an FD4 block each)
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          icplarrays = 1
    
          do idesc=1,ndesc
          
            ! cycle the loop if this is not my couple array
            if(desc(idesc)%rank /= mpi%rank) cycle

            FD4_L3_ASSERT(icplarrays>cpl%ncplarrays,'strange error')
            
            ! fill mycplarrays
            cpl%mycplarrays(icplarrays) = idesc

            ! set some convenience vars
            myabnd  => desc(idesc)%abnd
            idx = desc(idesc)%idx
            face = domain%vartab(idx)%facevar
            
            ! bounds of this non-existing block
            bbnd(1,:) = (/ domain%d(1)%bl(i) , domain%d(1)%bl(i+1) - 1 /)
            bbnd(2,:) = (/ domain%d(2)%bl(j) , domain%d(2)%bl(j+1) - 1 /)
            bbnd(3,:) = (/ domain%d(3)%bl(k) , domain%d(3)%bl(k+1) - 1 /)
            
            if(face==0) then
            !!! build datatype without face
            
              call fd4_couple_build_cplarray_type(desc(idesc), bbnd, domain%vartab(idx)%nbins, blocktype, err )
              FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
              if(blocktype /= MPI_DATATYPE_NULL) then
                ! save this type in the arraytypes array
                cpl%arraytypes(icplarrays, i, j, k, CPL_GET) = blocktype
                cpl%arraytypes(icplarrays, i, j, k, CPL_PUT) = blocktype
#ifdef FD4_COUPLE_DEBUG
                write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A)') &
                    'CREA ',idx,desc(idesc)%st,icplarrays,' p:',i,j,k,'  no face'
#endif              
              end if


            elseif(face>0 .and. face<4) then
            !!! build datatype with face

              ! additional cell plane if face var
              bbnd(face,2) = bbnd(face,2) + 1
              
              !! build the type for fd4_couple_put

              facetype = CPL_FACE
#ifdef FD4_COUPLE_FACE_ARRAYS_OVERLAP
              ! check if intersection is only one cell at the boundary of the block in face direction
              ! (if yes, the data will be provided by the neighbor block)
              if(myabnd(face,1)>=bbnd(face,2) .or.  &
                 myabnd(face,2)<=bbnd(face,1)    ) then
                 facetype = CPL_NO
              end if
#endif
              if(facetype == CPL_FACE) then
                call fd4_couple_build_cplarray_type(desc(idesc), bbnd, domain%vartab(idx)%nbins, blocktype, err )
                FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                if(blocktype /= MPI_DATATYPE_NULL) then
                  ! save this type in the arraytypes array
                  cpl%arraytypes(icplarrays, i, j, k, CPL_PUT) = blocktype
#ifdef FD4_COUPLE_DEBUG
                  write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A)') &
                      'CREA ',idx,desc(idesc)%st,icplarrays,' p:',i,j,k,'  face put'
#endif              
                end if
              end if

              !! build the type for fd4_couple_get
              
              facetype = CPL_NOFACE

              ! check if couple array ends at the first cell in face direction
              ! (if yes, the data will be provided by the neighbor block)
              if(desc(idesc)%abnd(face,2) <= bbnd(face,1)) then
                facetype = CPL_NO
              end if


#ifdef FD4_COUPLE_NEIGHBOR_FACE_CHECK
              pos(1) = i ; pos(2) = j ; pos(3) = k
              pos(face) = pos(face) + 1
              ! check if couple array ends at the additional face cell
              ! (saves transfer from the following block in this case)
              !      or
              ! neighbor block in face direction is not present in any rank
              if( desc(idesc)%abnd(face,2) == bbnd(face,2) .or. &
                  domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) == -1 ) then
                ! take the CPL_PUT type
                !  (it includes the additional face cell)
                facetype = CPL_FACE
                call fd4_couple_build_cplarray_type(desc(idesc), bbnd, domain%vartab(idx)%nbins, blocktype, err )
                FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                if(blocktype /= MPI_DATATYPE_NULL) then
                  ! save this type in the arraytypes array
                  cpl%arraytypes(icplarrays, i, j, k, CPL_GET) = blocktype
#ifdef FD4_COUPLE_DEBUG
                  write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A)') &
                      'CREA ',idx,desc(idesc)%st,icplarrays,' p:',i,j,k,'  face get CPL_FACE'
#endif              
                end if
              end if

#else
!FD4_COUPLE_NEIGHBOR_FACE_CHECK

              ! check if couple array ends at the additional face cell
              ! (saves transfer from the following block in this case)
              if(desc(idesc)%abnd(face,2) == bbnd(face,2)) then
                ! take the CPL_PUT type
                !  (it includes the additional face cell)
                facetype = CPL_FACE
                cpl%arraytypes(icplarrays, i, j, k, CPL_GET) = cpl%arraytypes(icplarrays, i, j, k, CPL_PUT)
#ifdef FD4_COUPLE_DEBUG
                write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A)') &
                    'CREA ',idx,desc(idesc)%st,icplarrays,' p:',i,j,k,'  face get CPL_FACE = put'
#endif              
              end if

#endif
!FD4_COUPLE_NEIGHBOR_FACE_CHECK

              ! build a type without the additional face cells
              if(facetype == CPL_NOFACE) then
                ! remove additional cell plane if face var
                bbnd(face,2) = bbnd(face,2) - 1
                call fd4_couple_build_cplarray_type(desc(idesc), bbnd, domain%vartab(idx)%nbins, blocktype, err )
                FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                if(blocktype /= MPI_DATATYPE_NULL) then
                  ! save this type in the arraytypes array
                  cpl%arraytypes(icplarrays, i, j, k, CPL_GET) = blocktype
#ifdef FD4_COUPLE_DEBUG
                  write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A)') &
                      'CREA ',idx,desc(idesc)%st,icplarrays,' p:',i,j,k,'  face get CPL_NOFACE'
#endif              
                end if
              end if
              

            end if ! face

            icplarrays = icplarrays + 1

          end do ! idesc          
          
        end do
      end do
    end do
    
    ! for each rank: calculate union of the bounds of all couple arrays
    allocate(cpl%rankbnd(3,2,0:mpi%nprocs-1))
    ! initialize with empty (negative) bounds
    do dest=0,mpi%nprocs-1
      cpl%rankbnd(1:3,1,dest) =  0
      cpl%rankbnd(1:3,2,dest) = -1
    end do
    ! add couple arrays to the bounds
    do idesc=1,ndesc
      dest = desc(idesc)%rank
      if( cpl%rankbnd(1,1,dest) > cpl%rankbnd(1,2,dest) ) then
        ! first block of rank dest
        cpl%rankbnd(1:3,1:2,dest) = desc(idesc)%abnd(1:3,1:2)
      else
        cpl%rankbnd(1:3,1,dest)=MIN(cpl%rankbnd(1:3,1,dest),desc(idesc)%abnd(1:3,1))
        cpl%rankbnd(1:3,2,dest)=MAX(cpl%rankbnd(1:3,2,dest),desc(idesc)%abnd(1:3,2))
      end if      
    end do
    
    err = 0

  end subroutine fd4_couple_create
  


  !! Build an MPI data type for a couple array. Internal routine.
  subroutine fd4_couple_build_cplarray_type(desc, bbnd, nbins, datatype, err )
    type(fd4_couple_desc), intent(in),target :: desc       !! description of couple array
    integer(i_k),          intent(in)        :: bbnd(3,2)  !! bounds of this (locally non-existing) block
    integer(i_k),          intent(in)        :: nbins      !! number of bins of the variable in desc
    integer,               intent(out)       :: datatype   !! the MPI data type
    integer(i_k),          intent(out)       :: err        !! error status: 0...ok
    
    integer(i_k) :: idx
    integer(i_k) :: sbnd(3,2),intersec(3,2),stride(1,2),ext(0:3)
    integer(i_k), pointer :: myabnd(:,:)

    err = -1
    datatype = MPI_DATATYPE_NULL

    ! set some convenience vars
    myabnd  => desc%abnd
    idx = desc%idx

    ! calculate intersection of block with array
    intersec(1:3,1)=MAX(bbnd(1:3,1),myabnd(1:3,1))
    intersec(1:3,2)=MIN(bbnd(1:3,2),myabnd(1:3,2))

    ! check if intersection is not null
    if(intersec(1,1)<=intersec(1,2) .and. &
       intersec(2,1)<=intersec(2,2) .and. &
       intersec(3,1)<=intersec(3,2) ) then

      if(nbins==1) then
        stride(1,1:2) = 1
        ext(0) = 1
        ext(1:3) = ubound(desc%array) - lbound(desc%array) + 1
      else
        stride(1,1) = desc%offset4
        stride(1,2) = desc%offset4 + nbins - 1
        stride(1,:) = stride(1,:) - lbound(desc%array4,1) + 1
        ext(0:3) = ubound(desc%array4) - lbound(desc%array4) + 1
      end if

      ! bounds of subset in array which respects to the current block, array content starts at given offset
      sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + desc%offset(1:3)
      sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + desc%offset(1:3)
      ! go, build the datatype for array section sbnd
      call fd4_comm_build_array_type(ext, stride, sbnd, datatype, err)
      FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')

    end if      

    err = 0
    
  end subroutine fd4_couple_build_cplarray_type
  


  
  !! Delete the couple context and all its memory.
  !!
  subroutine fd4_couple_delete(cpl, err)
    type(fd4_couple),  intent(inout)   :: cpl          !! couple context
    integer(i_k),      intent(out)     :: err          !! error status: 0...ok

    integer(i_k) :: i,j,k,icplarrays,dest
    integer, pointer :: cpl_put_type, cpl_get_type
    
    err = -1

    FD4_L1_ASSERT(.not. associated(cpl%domain),'cpl invalid')

    ! free fd4 mpi types
    do dest=0,cpl%domain%mpi%nprocs-1+cpl%toff
      call fd4_mpitype_free(cpl%commtype(dest),.true.,.false.)
    end do
    deallocate(cpl%commtype)

    ! free basic types for the couple array
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
          do icplarrays=1,cpl%ncplarrays
            cpl_put_type => cpl%arraytypes(icplarrays, i, j, k, CPL_PUT)
            cpl_get_type => cpl%arraytypes(icplarrays, i, j, k, CPL_GET)
            if( cpl_put_type /= MPI_DATATYPE_NULL ) then
              if( cpl_put_type == cpl_get_type) then
                ! cpl_put data type equals cpl_get data type (see fd4_couple_create)
                cpl_get_type = MPI_DATATYPE_NULL
              end if
              call MPI_Type_free( cpl_put_type, err)
              err = err - MPI_SUCCESS
              FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
            end if
            if( cpl_get_type /= MPI_DATATYPE_NULL ) then
              call MPI_Type_free( cpl_get_type, err)
              err = err - MPI_SUCCESS
              FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
            end if
          end do
        end do          
      end do
    end do
    
    deallocate(cpl%arraytypes)

    
    NULLIFY(cpl%domain)
    
    deallocate(cpl%rankbnd)
    deallocate(cpl%mycplarrays)
    deallocate(cpl%cpldescbnd)
    
    err = 0
  
  end subroutine fd4_couple_delete
  
    

  !! Put data to the framework. Each process contributes several blocks of
  !! data. $desc%abnd has to be set in all processes correctly. 
  !! You must supply the same $desc as in $fd4_couple_create, the only
  !! allowed change is the pointer to the data ($desc%array).
  !!
  !! *TODO*:
  !!  - see limitations above
  !!  - much more parameter checking
  !!  - check if we receive a cell twice (from different processes) ??
  !!
  subroutine fd4_couple_put(cpl, desc, err)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    type(fd4_couple_desc),intent(inout),target :: desc(:)      !! description of sender arrays
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: ndesc
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: index
    integer(i_k) :: idx, st, idesc, lb(4)
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,stride(1,2),fext(3),ext(3),bnd4th(2)
    real(r_k), pointer :: myarray(:,:,:), minptr
    integer(i_k), pointer :: myabnd(:,:)
    integer :: dest, blocktype, request
    integer(MPI_ADDRESS_KIND) :: disp
    type(fd4_block), pointer :: b
    type(fd4_iter)  :: iter
    logical :: is_min
    type(fd4_domain), pointer :: domain
    real(r_k), pointer :: mindisparray             ! pointer to couple array with min. address
    integer(MPI_ADDRESS_KIND) :: mindisp           ! address of couple array with min. address
    integer(i_k) :: icplarrays
    type(fd4_mpitype), pointer :: desttype
    
    err = -1
    FD4_L1_ASSERT(.not. associated(cpl%domain),'cpl invalid')
    
    ndesc = size(desc)

    domain => cpl%domain
    mpi => domain%mpi
    nullify(myabnd)
    mindisp = HUGE(1_MPI_ADDRESS_KIND)
    nullify(mindisparray)
    
    ! loop over local couple arrays:
    ! check params and get array with minimum displacement
    do icplarrays=1,cpl%ncplarrays
      idesc = cpl%mycplarrays(icplarrays)
      FD4_L3_ASSERT(desc(idesc)%rank/=mpi%rank,'mycplarrays not correct')
      idx = desc(idesc)%idx
      if( domain%vartab(idx)%nbins > 1 ) then
        myarray => desc(idesc)%array4(lbound(desc(idesc)%array4,1),:,:,:)
      else
        myarray => desc(idesc)%array
      end if
      if(.not. associated(myarray)) then
        FD4_L1_ERROR('desc%array not associated at '//i2s(idesc))
      end if
      ! find min displacement
      if( domain%vartab(idx)%nbins > 1 ) then
        lb(1:4) = lbound(desc(idesc)%array4)
        call MPI_Get_address( desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4))
        end if
      else
        lb(1:3) = lbound(desc(idesc)%array)
        call MPI_Get_address( desc(idesc)%array(lb(1),lb(2),lb(3)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => desc(idesc)%array(lb(1),lb(2),lb(3))
        end if
      end if
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! check array size for 4th dimension
      if( domain%vartab(idx)%nbins > 1 ) then
        if( lbound(desc(idesc)%array4,1) > desc(idesc)%offset4 ) then
          FD4_L1_ERROR('desc%offset4 too small at '//i2s(idesc))
        end if
        if( ubound(desc(idesc)%array4,1) < desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1 ) then
          FD4_L1_ERROR('desc%array4 too small at '//i2s(idesc))
        end if
      end if
      ! check array size for spatial dimensions
      fext = lbound(myarray)
      if( ANY( fext > desc(idesc)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      fext = ubound(myarray)
      if( ANY( fext < desc(idesc)%offset + desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
#endif
    end do


#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_put ',cpl%ncplarrays
#endif


    mpi%requests%pos = 0
  
    
    !
    !  1. send data
    !

    ! loop over all blocks of the domain
    ! to build an MPI datatype to send subsections
    ! from myarray to the blocks of other processes
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does not exist
          dest = domain%blkmap(i,j,k,domain%blkmap_now)
          if(dest == -1 ) cycle

          ! cycle if this is my block
          if(dest == mpi%rank) cycle
          
          ! loop over local couple arrays
          do icplarrays=1,cpl%ncplarrays
            
            ! get index of couple array in fd4_couple_desc
            idesc = cpl%mycplarrays(icplarrays)

            ! get the matching type from the array of pre-built datatypes
            blocktype = cpl%arraytypes(icplarrays, i, j, k, CPL_PUT)

            ! check if type exists
            if(blocktype /= MPI_DATATYPE_NULL) then

              index = INT( fd4_domain_ijk2if(i,j,k), i8k)   ! index for sorting
              
              ! either array or array4 must be associated!
              if(associated(desc(idesc)%array)) then
                lb(1:3) = lbound(desc(idesc)%array)
                call MPI_Get_address( desc(idesc)%array(lb(1),lb(2),lb(3)), disp, err)
              else
                lb(1:4) = lbound(desc(idesc)%array4)
                call MPI_Get_address( desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
              end if

              call fd4_mpitype_add(cpl%commtype(dest), index, 1, disp, blocktype)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I10,I7,I12,I7,I26)') &
                  'SEND ',desc(idesc)%idx,desc(idesc)%st,icplarrays,' p:',i,j,k,' id:',index,dest,blocktype,request,disp
#endif
            end if
            
          end do ! icplarrays
          
        end do
      end do
    end do
    
    
    ! send the data
    do dest=0,mpi%nprocs-1        

      desttype => cpl%commtype(dest)
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err, startdispl=mindisp)
        !call fd4_mpitype_commit(desttype, err, startdispl=0_MPI_ADDRESS_KIND)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(mindisparray),'strange error')
#ifdef FD4_COUPLE_DEBUG
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',desttype%types(i),desttype%indexes(i),desttype%displ(i),desttype%mindispl
        end do
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',sortmdt%types(i),sortmdt%displ(i),sortmdt%sort(i)
        end do
#endif
        ! send the data, use count as tag
        call MPI_Isend(mindisparray, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        !call MPI_Isend(MPI_BOTTOM, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(desttype,.false.,.false.)
#endif
      end if

    end do      

        

    
    !
    !  2. receive data
    !


    ! iterate over ranks
    do dest=0,mpi%nprocs-1
    
      ! calculate intersection of rectangular unions of
      !  - all my local blocks (including offset due to face variables in couple context)
      !  - all of rank dest's couple arrays
      ! if this intersection is null, we can safely skip
      ! further partition matching checks of this rank
      if( .not. section_intersect(domain%blockunion, cpl%rankbnd(1:3,1:2,dest), cpl%faceoffset) ) cycle
      
      desttype => cpl%commtype(dest+cpl%toff)
    
      do idesc=cpl%cpldescbnd(1,dest),cpl%cpldescbnd(2,dest)

        ! cycle the loop if this couple array does not belong to rank dest
        if(desc(idesc)%rank /= dest) cycle

        ! set some convenience vars
        idx = desc(idesc)%idx
        st  = desc(idesc)%st
        face = domain%vartab(idx)%facevar

        ! stride for 4th dimension
        stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
        
        fext = 0
        if(face>0 .and. face<4) then
          fext(face) = fext(face) + 1
        end if      

        ! iterate over existing blocks and create a commtype
        call fd4_iter_init(domain, iter)
        do while(associated(iter%cur))
          b => iter%cur

          bbnd = b%bnd

          if(face>0 .and. face<4) then
            ! additional cell plane if face var
            bbnd(face,2) = bbnd(face,2) + 1
          end if

          ! calculate intersection of block with array
          intersec(1:3,1)=MAX(bbnd(1:3,1),desc(idesc)%abnd(1:3,1))
          intersec(1:3,2)=MIN(bbnd(1:3,2),desc(idesc)%abnd(1:3,2))

#ifdef FD4_COUPLE_FACE_ARRAYS_OVERLAP
          if(face>0 .and. face<4) then
            ! check if intersection is only one cell at the boundary of the block in face direction
            ! (if yes, the data will be provided by the neighbor block)
            if(intersec(face,1)>=bbnd(face,2) .or.  &
               intersec(face,2)<=bbnd(face,1)    ) then
               ! in this case, we just set the intersection to null, to force the next if to be false
               intersec(1,2) = intersec(1,1) - 1
            end if
          end if          
#endif

          ! check if intersection is not null
          if(intersec(1,1)<=intersec(1,2) .and. &
             intersec(2,1)<=intersec(2,2) .and. &
             intersec(3,1)<=intersec(3,2) ) then

            ! bounds of subset in block which respects to the intersection
            bbnd(1:3,1) = intersec(1:3,1) - b%bnd(1:3,1) + 1
            bbnd(1:3,2) = intersec(1:3,2) - b%bnd(1:3,1) + 1
             
            if(dest /= mpi%rank) then
              ! get MPI type for the block + variable
              ext = b%ext + fext
              call fd4_couple_blocktype(cpl%domain, idx, st, ext, bbnd, blocktype)
              call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
              ! index for sorting
              index = INT( fd4_domain_c2if(b%pos), i8k)   ! index for sorting
              call fd4_mpitype_add( desttype, index, 1, disp, blocktype, is_min)
              if(is_min) minptr => b%sdata(face)%l(1,1,1,1)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,I10,I7,I12,I7,I26)') &
                  'RECV ',idx,st,'',' p:',b%pos,' id:',index,dest,blocktype,request,disp
                  !,loc(b%sdata(face)%l(1,1,1,1))
#endif
            else
              ! copy local array section to block instead of sending msg to myself
              myabnd  => desc(idesc)%abnd
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              ! either array (nbins=1) or array4 (nbins>1) must be associated!
              if(domain%vartab(idx)%nbins == 1) then
                b%sdata(face)%l(stride(1,1),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) &
                  = desc(idesc)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2))
              else
                bnd4th(1) = desc(idesc)%offset4
                bnd4th(2) = desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1
                b%sdata(face)%l(stride(1,1):stride(1,2),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) &
                  = desc(idesc)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2))
              end if
#ifdef FD4_COUPLE_DEBUG
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5)') &
                  'COPY ',idx,st,'',' p:',b%pos
#endif
            end if ! dest /= mpi%rank

          end if ! intersection

          call fd4_iter_next(iter)
        end do ! blocks
      
      end do ! desc

      ! now start receiving the data, use count as tag
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err)
        !call fd4_mpitype_commit(desttype, err, 0_MPI_ADDRESS_KIND)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
#ifdef FD4_COUPLE_DEBUG
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',desttype%types(i),desttype%indexes(i),desttype%displ(i),desttype%mindispl
        end do
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',sortmdt%types(i),sortmdt%displ(i),sortmdt%sort(i)
        end do
#endif
        ! recv the data
        call MPI_Irecv(minptr, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err)
        !call MPI_Irecv(MPI_BOTTOM, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(desttype,.false.,.false.)
#endif
      end if
      
    end do ! ranks


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
#ifdef FD4_COUPLE_DEBUG
       write(300+mpi%rank,*) 'WAIT ',mpi%requests%pos
#endif
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    


#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    do dest=0,mpi%nprocs-1+cpl%toff
      if(cpl%commtype(dest)%count>0)  call fd4_mpitype_free(cpl%commtype(dest),.false.,.false.)
    end do
#endif

    err = 0    
  
  end subroutine fd4_couple_put
  
  

  !! Get data from the framework. Each process asks for several blocks of
  !! data. $desc%abnd has to be set in all processes correctly. 
  !!
  !! *Additional limitation for face variables*: The upper face cell plane of the blocks is
  !! just cut and not transferred to the couple arrays!
  !!
  !! *TODO*:
  !!  - see limitations above
  !!  - much more parameter checking
  !!
  subroutine fd4_couple_get(cpl, desc, err, opt_novnull)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    type(fd4_couple_desc),intent(inout),target :: desc(:)      !! description of receiver arrays
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    logical, optional,    intent(in)           :: opt_novnull  !! do not initalize the receiver arrays with vnull
    
    integer(i_k) :: ndesc
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: index
    integer(i_k) :: idx, st, idesc, icplarrays, lb(4)
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,stride(1,2),fext(3),ext(3),bnd4th(2)
    real(r_k), pointer :: myarray(:,:,:), mindisparray, minptr
    integer(i_k), pointer :: myabnd(:,:)
    integer :: dest, blocktype, request
    integer(MPI_ADDRESS_KIND) :: disp, mindisp
    type(fd4_block), pointer :: b
    type(fd4_iter)  :: iter
    logical :: is_min
    type(fd4_domain), pointer :: domain
    logical :: novnull
#ifdef FD4_COUPLE_NEIGHBOR_FACE_CHECK
    integer(i_k) :: pos(3)
#endif
    type(fd4_mpitype), pointer :: desttype

    err = -1
    FD4_L1_ASSERT(.not. associated(cpl%domain),'cpl invalid')

    ndesc = size(desc)

    domain => cpl%domain
    mpi => domain%mpi
    nullify(mindisparray)
    mindisp = HUGE(1_MPI_ADDRESS_KIND)
    
    novnull = .false.
    if(present(opt_novnull)) novnull = opt_novnull


    ! loop over local couple arrays:
    ! check params and get array with minimum displacement
    ! and initialize coupling arrays with vnull
    do icplarrays=1,cpl%ncplarrays
      idesc = cpl%mycplarrays(icplarrays)
      FD4_L3_ASSERT(desc(idesc)%rank/=mpi%rank,'mycplarrays not correct')
      idx = desc(idesc)%idx
      if( domain%vartab(idx)%nbins > 1 ) then
        myarray => desc(idesc)%array4(lbound(desc(idesc)%array4,1),:,:,:)
      else
        myarray => desc(idesc)%array
      end if
      if(.not. associated(myarray)) then
        FD4_L1_ERROR('desc%array not associated at '//i2s(idesc))
      end if
      ! find min displacement
      if( domain%vartab(idx)%nbins > 1 ) then
        lb(1:4) = lbound(desc(idesc)%array4)
        call MPI_Get_address( desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4))
        end if
      else
        lb(1:3) = lbound(desc(idesc)%array)
        call MPI_Get_address( desc(idesc)%array(lb(1),lb(2),lb(3)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => desc(idesc)%array(lb(1),lb(2),lb(3))
        end if
      end if
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! check array size for 4th dimension
      if( domain%vartab(idx)%nbins > 1 ) then
        if( lbound(desc(idesc)%array4,1) > desc(idesc)%offset4 ) then
          FD4_L1_ERROR('desc%offset4 too small at '//i2s(idesc))
        end if
        if( ubound(desc(idesc)%array4,1) < desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1 ) then
          FD4_L1_ERROR('desc%array4 too small at '//i2s(idesc))
        end if
      end if
      ! check array size for spatial dimensions
      fext = lbound(myarray)
      if( ANY( fext > desc(idesc)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      fext = ubound(myarray)
      if( ANY( fext < desc(idesc)%offset + desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
#endif
      ! initialize array with vnull for specified variable
      if(.not. novnull) then
        sbnd(1:3,1) = desc(idesc)%offset
        sbnd(1:3,2) = desc(idesc)%offset + desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1)
        if( domain%vartab(idx)%nbins > 1 ) then
          bnd4th(1) = desc(idesc)%offset4
          bnd4th(2) = desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1
          desc(idesc)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
            = domain%vartab(idx)%vnull
        else
          desc(idesc)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) & 
            = domain%vartab(idx)%vnull
        end if
      end if
    end do

    mpi%requests%pos = 0
 
#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_get ',cpl%ncplarrays
#endif

    
    !
    !  1. send data
    !


    ! iterate over ranks
    do dest=0,mpi%nprocs-1

      ! calculate intersection of rectangular unions of
      !  - all my local blocks (including offset due to face variables in couple context)
      !  - all of rank dest's couple arrays
      ! if this intersection is null, we can safely skip
      ! further partition matching checks of this rank
      if( .not. section_intersect(domain%blockunion, cpl%rankbnd(1:3,1:2,dest), cpl%faceoffset) ) cycle

      desttype => cpl%commtype(dest+cpl%toff)

      do idesc=cpl%cpldescbnd(1,dest),cpl%cpldescbnd(2,dest)

        ! cycle the loop if this couple array does not belong to rank dest
        if(desc(idesc)%rank /= dest) cycle

        ! set some convenience vars
        idx = desc(idesc)%idx
        st  = desc(idesc)%st
        face = domain%vartab(idx)%facevar

        ! stride for 4th dimension
        stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
        
        fext = 0
        if(face>0 .and. face<4) then
          fext(face) = fext(face) + 1
        end if      

        ! iterate over existing blocks and create a commtype
        call fd4_iter_init(domain, iter)
        do while(associated(iter%cur))
          b => iter%cur

          bbnd = b%bnd

          if(face>0 .and. face<4) then
            ! additional cell plane if face var
            !     and
            !   couple array ends at the additional cell
            !   (saves transfer from the following block in this case)
            !      or
            !   neighbor block in face direction is not present in any rank
#ifdef FD4_COUPLE_NEIGHBOR_FACE_CHECK
            pos(:) = b%pos(:)
            pos(face) = pos(face) + 1
            if( desc(idesc)%abnd(face,2) == bbnd(face,2)+1 .or. &
                domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) == -1 ) then
#else
            if( desc(idesc)%abnd(face,2) == bbnd(face,2)+1 ) then
#endif
              bbnd(face,2) = bbnd(face,2) + 1
            end if
          end if

          ! calculate intersection of block with array
          intersec(1:3,1)=MAX(bbnd(1:3,1),desc(idesc)%abnd(1:3,1))
          intersec(1:3,2)=MIN(bbnd(1:3,2),desc(idesc)%abnd(1:3,2))

          if(face>0 .and. face<4) then
            ! check if intersection is only one cell at the boundary of the block in face direction
            ! (if yes, the data will be provided by the neighbor block)
            if(desc(idesc)%abnd(face,2) <= bbnd(face,1)) then
               ! in this case, we just set the intersection to null, to force the next if to be false
               intersec(1,2) = intersec(1,1) - 1
            end if
          end if          

          ! check if intersection is not null
          if(intersec(1,1)<=intersec(1,2) .and. &
             intersec(2,1)<=intersec(2,2) .and. &
             intersec(3,1)<=intersec(3,2) ) then

            ! bounds of subset in block which respects to the intersection
            bbnd(1:3,1) = intersec(1:3,1) - b%bnd(1:3,1) + 1
            bbnd(1:3,2) = intersec(1:3,2) - b%bnd(1:3,1) + 1
             
            if(dest /= mpi%rank) then
              ! get MPI type for the block + variable
              ext = b%ext + fext
              call fd4_couple_blocktype(cpl%domain, idx, st, ext, bbnd, blocktype)
              call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
              ! index for sorting
              index = INT( fd4_domain_c2if(b%pos), i8k)   ! index for sorting
              call fd4_mpitype_add( desttype, index, 1, disp, blocktype, is_min)
              if(is_min) minptr => b%sdata(face)%l(1,1,1,1)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,I10,I7,I12,I7)') &
                  'SEND ',idx,st,'',' p:',b%pos,' id:',index,dest,blocktype,request
#endif
            else
              ! copy local array section to block instead of sending msg to myself
              myabnd  => desc(idesc)%abnd
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              ! either array (nbins=1) or array4 (nbins>1) must be associated!
              if(domain%vartab(idx)%nbins == 1) then
                desc(idesc)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
                  = b%sdata(face)%l(stride(1,1),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2))
              else
                bnd4th(1) = desc(idesc)%offset4
                bnd4th(2) = desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1
                desc(idesc)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
                  = b%sdata(face)%l(stride(1,1):stride(1,2),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2))
                
              end if
#ifdef FD4_COUPLE_DEBUG
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,I10,I7)') &
                  'COPY ',idx,st,'',' p:',b%pos
#endif
            end if ! dest /= mpi%rank

          end if ! intersection

          call fd4_iter_next(iter)
        end do ! blocks


      
      end do ! desc

      
      ! now start sending the data, use count as tag
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        ! recv the data
        call MPI_Isend(minptr, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(desttype,.false.,.false.)
#endif
      end if
      

    end do ! ranks


    
    !
    !  2. receive data
    !

    ! loop over all blocks of the domain
    ! to build an MPI datatype to recv subsections
    ! from the blocks of other processes to my couple arrays
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does not exist
          dest = domain%blkmap(i,j,k,domain%blkmap_now)
          if(dest == -1 ) cycle

          ! cycle if this is my block
          if(dest == mpi%rank) cycle

          ! loop over local couple arrays
          do icplarrays=1,cpl%ncplarrays
          
            ! get index of couple array in fd4_couple_desc
            idesc = cpl%mycplarrays(icplarrays)

            ! get the matching type from the array of pre-built datatypes
            blocktype = cpl%arraytypes(icplarrays, i, j, k, CPL_GET)

            ! check if type exists
            if(blocktype /= MPI_DATATYPE_NULL) then

              index = INT( fd4_domain_ijk2if(i,j,k), i8k)   ! index for sorting

              ! either array or array4 must be associated!
              if(associated(desc(idesc)%array)) then
                lb(1:3) = lbound(desc(idesc)%array)
                call MPI_Get_address( desc(idesc)%array(lb(1),lb(2),lb(3)), disp, err)
              else
                lb(1:4) = lbound(desc(idesc)%array4)
                call MPI_Get_address( desc(idesc)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
              end if

              call fd4_mpitype_add(cpl%commtype(dest), index, 1, disp, blocktype)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I10,I7,I12,I7)') &
                  'RECV ',desc(idesc)%idx,desc(idesc)%st,icplarrays,' p:',i,j,k,' id:',index,dest,blocktype,request
#endif

            end if

          end do ! icplarrays
          
        end do
      end do
    end do
    
    
    ! recv the data
    do dest=0,mpi%nprocs-1        

      desttype => cpl%commtype(dest)
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err, startdispl=mindisp)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(mindisparray),'strange error')
        ! recv  the data, use count as tag
        call MPI_Irecv(mindisparray, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(desttype,.false.,.false.)
#endif
      end if

    end do      


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
#ifdef FD4_COUPLE_DEBUG
       write(300+mpi%rank,*) 'WAIT ',mpi%requests%pos
#endif
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    


#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    do dest=0,mpi%nprocs-1+cpl%toff
      if(cpl%commtype(dest)%count>0)  call fd4_mpitype_free(cpl%commtype(dest),.false.,.false.)
    end do
#endif

    err = 0    
  
  end subroutine fd4_couple_get





  !! Get an MPI data type for the variable described by ($idx, $st), a block
  !! of size $ext, and a spatial subsection within the block described
  !! by $bbnd.
  !! If an MPI data type with same properties has already been created,
  !! this type will be re-used.
  !! 
  !! Internal routine. Used in $fd4_couple_put and $fd4_couple_get.
  subroutine fd4_couple_blocktype(domain, idx, st, ext, bnd, blocktype)
    type(fd4_domain), intent(inout)  :: domain  !! the domain
    integer(i_k), intent(in)  :: idx            !! variable index
    integer(i_k), intent(in)  :: st             !! time step index
    integer(i_k), intent(in)  :: ext(3)         !! block extents
    integer(i_k), intent(in)  :: bnd(3,2)       !! bound of spatial subset
    integer,      intent(out) :: blocktype      !! MPI data type for the block
    
    integer(i8k) :: index
    integer(i_k) :: err, stride(1,2), ext4(0:3)
    logical :: blocktype_present
    
    ! get index of the type in domain-global cpltypestore
    call fd4_domain_block_encode(domain%cplblkenc, idx, st, ext(1:3), bnd, index)

    ! check if the type is present already
    blocktype = rbtree_int8_get(domain%cpltypestore, index, found = blocktype_present)

#ifdef FD4_COUPLE_DEBUG
    write(300+domain%mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,3(I5),A,3(I5),L5,I16)') &
        'BLTY ',idx,st,'',' ext:',ext,'  bnd:',bnd(:,1),'-',bnd(:,2), blocktype_present, index
#endif

    if(.not. blocktype_present) then
      stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
      ext4(0) = domain%varinfo%field_len(domain%vartab(idx)%facevar)
      ext4(1:3) = ext(1:3)
      VT_USER_START('FD4_COUPLE_PUT.build_array_type')
      call fd4_comm_build_array_type(ext4, stride, bnd, blocktype, err)
      VT_USER_END('FD4_COUPLE_PUT.build_array_type')
      FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
      ! put type to blockstore
      call rbtree_int8_insert(domain%cpltypestore, index, blocktype, err)
      FD4_L3_ASSERT(err/=0,'rbtree_int_insert failed')
    end if

  end subroutine fd4_couple_blocktype







  !! Pretend putting data to the framework and mark all framework blocks
  !! which need to be created to fully capture all threshold variables.
  !!
  !! Use this in a 3-stage coupling process:
  !!  - $fd4_couple_mark_blocks to mark required blocks
  !!  - $fd4_balance_readjust to ensure these blocks are created and balanced
  !!  - $fd4_couple_put to finally put the data to fd4
  !! You must use the same parameters for $fd4_couple_mark_blocks as for $fd4_couple_put.
  !!
  !! *TODO*:
  !!  - apply performance optimizations used in $fd4_couple_put and $fd4_couple_get here
  !!  - see limitations above
  !!  - much more parameter checking
  !!
  subroutine fd4_couple_mark_blocks(cpl, desc, err)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    type(fd4_couple_desc),intent(in),target    :: desc(:)      !! description of sender arrays
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: ndesc
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: idx, idesc
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,bnd4th(2)
    real(r_k), pointer :: myarray(:,:,:)
    integer(i_k), pointer :: myabnd(:,:)
    type(fd4_domain), pointer :: domain
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: fext(3)
#endif

    err = -1
    FD4_L1_ASSERT(.not. associated(cpl%domain),'cpl invalid')

    ndesc = size(desc)
    
    domain => cpl%domain
    mpi => domain%mpi
    nullify(myabnd)
    intersec(:,1) = 0
    intersec(:,2) = -1

    ! check params
    ! find couple array with min address
    do idesc=1,ndesc
      idx = desc(idesc)%idx
      if(desc(idesc)%rank == mpi%rank) then
        myarray => desc(idesc)%array
        if( domain%vartab(idx)%nbins > 1 ) myarray => desc(idesc)%array4(lbound(desc(idesc)%array4,1),:,:,:)
        if(.not. associated(myarray)) then
          FD4_L1_ERROR('desc%array not associated at '//i2s(idesc))
        end if
#if ( FD4_VERBOSE_LEVEL >= 3 )
        ! check array size for 4th dimension
        if( domain%vartab(idx)%nbins > 1 ) then
          if( lbound(desc(idesc)%array4,1) > desc(idesc)%offset4 ) then
            FD4_L1_ERROR('desc%offset4 too small at '//i2s(idesc))
          end if
          if( ubound(desc(idesc)%array4,1) < desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1 ) then
            FD4_L1_ERROR('desc%array4 too small at '//i2s(idesc))
          end if
        end if
        ! check array size for spatial dimensions
        fext = lbound(myarray)
        if( ANY( fext > desc(idesc)%offset ) ) then
          FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
        end if
        fext = ubound(myarray)
        if( ANY( fext < desc(idesc)%offset + desc(idesc)%abnd(1:3,2) - desc(idesc)%abnd(1:3,1) ) ) then
          FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
        end if
#endif
      end if
    end do


    
    ! loop over all blocks of the domain

    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does exist
          !    actually, this is a bit inaccurate!! fd4_balance_readjust
          !    may decide to delete a block which we may found to be
          !    required - if we would not cycle now. but running
          !    this check for all blocks is extremely costly for high
          !    block numbers.
          if(domain%blkmap(i,j,k,domain%blkmap_now) > -1 ) cycle

          idesc_loop: do idesc=1,ndesc
          
            ! cycle the loop if this is not my couple array
            if(desc(idesc)%rank /= mpi%rank) cycle

            idx = desc(idesc)%idx

            ! cycle the loop if this var has no threshold
            if(domain%vartab(idx)%vthres == FD4_NOTHRES) cycle

            ! set some convenience vars
            myabnd  => desc(idesc)%abnd
            face = domain%vartab(idx)%facevar
            
            ! bounds of this non-existing block
            bbnd(1,:) = (/ domain%d(1)%bl(i) , domain%d(1)%bl(i+1) - 1 /)
            bbnd(2,:) = (/ domain%d(2)%bl(j) , domain%d(2)%bl(j+1) - 1 /)
            bbnd(3,:) = (/ domain%d(3)%bl(k) , domain%d(3)%bl(k+1) - 1 /)

            if(face>0 .and. face<4) then
              ! additional cell plane if face var
              bbnd(face,2) = bbnd(face,2) + 1
            end if

            ! calculate intersection of block with array
            intersec(1:3,1)=MAX(bbnd(1:3,1),myabnd(1:3,1))
            intersec(1:3,2)=MIN(bbnd(1:3,2),myabnd(1:3,2))

            ! check if intersection is not null
            if(intersec(1,1)<=intersec(1,2) .and. &
               intersec(2,1)<=intersec(2,2) .and. &
               intersec(3,1)<=intersec(3,2) ) then

              ! bounds of subset in array which respects to the current block, array content starts at given offset
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + desc(idesc)%offset(1:3)
              !print *,mpi%rank,':',sbnd(1:3,1),' - ',sbnd(1:3,2)

              ! either array or array4 must be associated!
              if(associated(desc(idesc)%array)) then
                ! if any of the values in the intersection is larger than vthres, we mark this block
                if(   ANY( abs(desc(idesc)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2))) &
                    > domain%vartab(idx)%vthres ) ) then
                  call stack_push( domain%markedblocks, fd4_domain_ijk2if(i,j,k) )
                  ! we marked this block, so we need not to evaluate the other descriptions at this pos
                  exit idesc_loop
                end if
              else
                ! if any of the values in the intersection is larger than vthres, we mark this block
                bnd4th(1) = desc(idesc)%offset4
                bnd4th(2) = desc(idesc)%offset4 + domain%vartab(idx)%nbins - 1
                if(   ANY( abs(desc(idesc)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)))&
                    > domain%vartab(idx)%vthres ) ) then
                  call stack_push( domain%markedblocks, fd4_domain_ijk2if(i,j,k) )
                  ! we marked this block, so we need not to evaluate the other descriptions at this pos
                  exit idesc_loop
                end if
              end if

            end if ! intersection

          end do idesc_loop ! idesc          
          
        end do
      end do
    end do
    
    ! merge the local stacks to one global
    call fd4_couple_stack_global_merge(domain%mpi, domain%markedblocks, err)
    FD4_L3_ASSERT(err/=0,'fd4_couple_stack_global_merge failed')


    ! important: set the flag, that we have marked blocks,
    !            so fd4_balance_readjust knows this
    domain%blocks_are_marked = ( domain%markedblocks%pos > 0 )
    
    err = 0    
  
  end subroutine fd4_couple_mark_blocks




  !! Merge the process-local stacks $st to one global stack in every
  !! process.
  !!
  !! Duplicates will be removed, the stack will be sorted.
  !! It is assumed that the process local stacks are already sorted.
  !!
  !! Internal routine.
  !!
  subroutine fd4_couple_stack_global_merge(mpi, st, err)
    type(fd4_mpi),     intent(inout) :: mpi               !! the domain's mpi context
    type(stack),       intent(inout) :: st                !! the stack 
    integer(i_k),      intent(out)   :: err               !! error status: 0...ok
    
    integer :: count, i
    type(stack) :: st_global
    type(rbtree_int) :: intree
    type(rbtree_int_iterator) :: initer
    integer(i_k) :: p, stat
    type(stack) :: st_sort

    err = -1   

    !do i=1,st%pos
    !  write(300+mpi%rank,'(I20,I5,I5,I5)') st%a(i),fd4_domain_i2cf(st%a(i))
    !end do
    
    !! tell all other processes how many entries my local stack3 has
    call MPI_Allgather( st%pos, 1, MPI_I_K, mpi%recvcnt, 1, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
    
    !! let's see if there's anything, if no, we can leave this routine
    count = SUM(mpi%recvcnt)
    if(count == 0) then
      return
    end if
        
    !! now collect the contents of the local stacks
    
    !! set receive displacements for MPI_Allgatherv
    mpi%recvdis(0) = 0
    do i=0,mpi%nprocs-2
      mpi%recvdis(i+1) = mpi%recvdis(i) + mpi%recvcnt(i)
    end do

    !! allocate memory in the global stack
    call stack_create(st_global, count)
    st_global%pos = count

    !! go! receive directly into the new created st_global
    call MPI_Allgatherv (       st%a,       st%pos,                MPI_I_K, &
                         st_global%a,  mpi%recvcnt,   mpi%recvdis, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
    
    call stack_move(st_global, st)

    !! sort, i.e. merge the sorted parts of the st_global to one sorted stack st_sort

    call stack_create(st_sort, count)
    ! senddis contains the end indices of the procs in st
    mpi%senddis(:) = mpi%recvdis(:) + mpi%recvcnt(:)
    ! recvdis contains the start indices of the procs in st (see above)
    mpi%recvdis(:) = mpi%recvdis(:) + 1
    
    ! initialize the tree of integers (maps block idx to rank, sorting by block idx)
    !  holds first block idx of each rank, automatic sorting to find
    !  smallest block idx
    do p=0,mpi%nprocs-1
      if(mpi%recvdis(p)>mpi%senddis(p)) cycle
      do
        call rbtree_int_insert(intree, st%a( mpi%recvdis(p) ), p, stat )
        if(stat==0) exit
        mpi%recvdis(p) = mpi%recvdis(p) + 1
        if(mpi%recvdis(p)>mpi%senddis(p)) exit
      end do
    end do
    
    !call rbtree_int_print(intree)
    
    do while(intree%length>0)
      ! first element = smallest remaining block index 
      p = rbtree_int_iterator_first(initer,intree)
      ! push this value to the sorted stack
      st_sort%pos = st_sort%pos + 1
      !print *,st_sort%pos,'of',size(st_sort%a),p,initer%cur%key
      FD4_L3_ASSERT(st_sort%pos>size(st_sort%a),'st_sort too small!')
      FD4_L3_ASSERT(initer%cur%key/=st%a(mpi%recvdis(p)),'tree or stack corrupt!')
      st_sort%a(st_sort%pos) = initer%cur%key
      ! delete this element, this invalidates the iterator
      p = rbtree_int_delete(intree, initer%cur%key)
      ! increase start index
      mpi%recvdis(p) = mpi%recvdis(p) + 1
      ! put the min block idx of this rank to the tree
      if(mpi%recvdis(p)>mpi%senddis(p)) cycle
      do
        call rbtree_int_insert(intree, st%a( mpi%recvdis(p) ), p, stat )
        if(stat==0) exit
        ! need to take next item, because of duplicates
        mpi%recvdis(p) = mpi%recvdis(p) + 1
        if(mpi%recvdis(p)>mpi%senddis(p)) exit
      end do
    end do

    ! this is actually not necessary
    call rbtree_int_clear(intree)

    call stack_move(st_sort, st)
   
    !do i=1,st%pos
    !  write(200+mpi%rank,'(I20,I5,I5,I5)') st%a(i),fd4_domain_i2cf(st%a(i))
    !end do
    
    err = 0

  end subroutine fd4_couple_stack_global_merge
  

end module fd4_couple_mod
